
//******************************************************************************
//  FILE..........: CASLINK.C
//  COPYRIGHT.....: Copyright (C) 1999 Alexey Podrezov
//  VERSION.......: 1.0b
//  DESCRIPTION...: Cassette Interface Emulation utility
//  NOTES.........: Encodes/decodes files to be transfered through MSX cassette interface
//  HISTORY.......:
//  PERSON          DATE         COMMENT
//  --------------- ------------ -----------------------------------------------
//  Podrezov        20/11/1998   Initial version created

//******************************************************************************


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <conio.h>

#define STARTBIT 0
#define STOPBIT 1
#define SILENCE 0x80
#define HIGHAMPLITUDE 0xff
#define LOWAMPLITUDE 0
#define MAXSIGNALLEVEL 100
#define CASFILENAMELENGTH 6
#define WAVHEADEROFFSETVALUE 0x77
#define MAXENCINPUTFILELENGTH 33000

int WriteDataByte(char);
void ExitOnError(int ErrorNum);
void CleanUp(void);
void DecodeFile(void);
void EncodeFile(void);

typedef unsigned int WORD;
typedef unsigned long DWORD;

char InputFileName[13], OutputFileName[13];
unsigned int WavSampleRate, SignalLevel, MSXInterfaceSpeed, LongHeaderLength, ShortHeaderLength, SampleScale, SpeedScale;
FILE *InputFileHandle, *OutputFileHandle;
char BitStream[11], *InputMemPointer, *OutputMemPointer, FirstPauseLength, MiddlePauseLength, FinalPauseLength, SilenceByte, EncodeDecode, FileType;
size_t InputFileLength, OutputFileLength, BytesRead, BytesWritten, FileOffset, ExtraBytes;
struct {
	char WavHeaderID[4];	// 'RIFF'
   DWORD SampleLength;
   char MMFileType[4];	// 'WAVE'
  	char FormatID[4];	// 'fmt '
   DWORD FormatSize;
   WORD FormatTag;
   WORD NumChannels;
   DWORD SamplesPerSec;
   DWORD BytesPerSec;
	WORD Dummy;
   WORD BitsPerSample;
	char DataID[4];	// 'data'
   DWORD PureSampleLength;
   } WavHeader;

char BasicFileHeaderID[10] = {0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3, 0xd3};
char BinFileHeaderID[10] = {0xd0, 0xd0, 0xd0, 0xd0, 0xd0, 0xd0, 0xd0, 0xd0, 0xd0, 0xd0};
char DataFileHeaderID[10] = {0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea, 0xea};
char NameBuffer[CASFILENAMELENGTH], ch;
char BitEncodingLength, FileHeaderIDLength;
size_t i;
DWORD j;


// Patterns of 1200Hz and 2400Hz waves - samples for 11025Hz file. if bigger sampling rate is used, multiply by SampleScale variable
unsigned char ZeroBit[2][10] = {
	{LOWAMPLITUDE, LOWAMPLITUDE, LOWAMPLITUDE, LOWAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE},
	{LOWAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE}
   };
unsigned char SetBit[2][10] = {
	{LOWAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE, HIGHAMPLITUDE},
   {LOWAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE, LOWAMPLITUDE, HIGHAMPLITUDE}
   };

char ErrorList[][80] = {
   " ERROR: Incorrect parameter passed to command line...\n\n",								// 1
   " ERROR: Insufficient command line information. Need a key and a filename...\n\n",	// 2
   " ERROR: Incorrect value specified...\n\n",														// 3
   " ERROR: Failed to allocate memory for output buffer...\n\n", 								// 4
   " ERROR: Failed to allocate memory for input buffer...\n\n", 								// 5
   " ERROR: Failed to open input file...\n\n", 														// 6
   " ERROR: Failed to manipulate input file pointer...\n\n", 									// 7
   " ERROR: Input file can't be processed...\n\n", 												// 8
   " ERROR: Failed to read input file...\n\n", 														// 9
   " ERROR: Input file format is not supported...\n\n", 											// 10
   " CANCEL: Overwrite request canceled by user...\n\n", 										// 11
   " ERROR: Failed to create output file...\n\n", 		 											// 12
   " ERROR: Failed to write to output file...\n\n", 		 										// 13
   ""
   };


// Main routine ----------------------------------------------------------------

int main(int argc, char *argv[])
{

   // Print title
   printf("\nCasLink Version 1.0b  Copyright (c) 1999 Alex Podrezov. All rights reserved.\n");
   printf("\n This utility encodes/decodes files to be transfered through MSX cassette");
   printf("\n interface by any sound I/O capable hardware - PC sound card or CD player.\n");
   printf("\n Supported input formats : BAS (ascii/tokenized), BIN (16/32k), ROM (16/32k)");
   printf("\n Supported output formats: BAS (ascii/tokenized), BIN (16/32k), ROM (16/32k)\n");

	//Output help hint
   if (argc == 1)
   {
   	printf("\n To get help on command line options use /? key...\n\n");
      return 0;
      }

   // Check command line parameters
	if (argc != 3 && *(char*)(argv[1]+1) != '?') ExitOnError(2);
   if (strlen(argv[1]) != 2 || strlen(argv[2]) > 12) ExitOnError(1);
   if (*(char*)(argv[1]) != '/') ExitOnError(1);
   else
   {
   	if (*(char*)(argv[1] + 1) == '?')
      {
		   printf("\n");
		   printf("\n Usage: CasLink <key> <filename>\n");
		   printf("\n  Keys: <filename> - name of input/output file (12 characters max);");
		   printf("\n        </E> - encode basic or binary file into WAV sound file;");
		   printf("\n        </D> - decode basic or binary file from WAV sound file;");
		   printf("\n        </?> - this help screen.");
		   printf("\n");
		   printf("\n  Note: File name encoded to WAV file will be in capital letters.");
		   printf("\n\n");
         return 0;
         }
      else if (*(char*)(argv[1] + 1) == 'E' || *(char*)(argv[1] + 1) == 'e') EncodeDecode = 1;	// Encode to WAV
      else if (*(char*)(argv[1] + 1) == 'D' || *(char*)(argv[1] + 1) == 'd') EncodeDecode = 2;	// Decode from WAV
      else ExitOnError(1);

      // Uppercase the string and copy input/output file name to buffers
      for (i = 0; i < strlen(argv[2]); i++)
      {
      	*(char*)(argv[2]+i) = (char)toupper(*(char*)(argv[2]+i));
			(char)InputFileName[i] = *(char*)(argv[2]+i);
         }
      (char)InputFileName[i] = '\0';
		for (i = 0; i < 8; i++)
      {
      	if ((char)InputFileName[i] == '.') break;
			(char)OutputFileName[i] = (char)InputFileName[i];
         }
      (char)OutputFileName[i] = '\0';
      if (EncodeDecode == 1) strcat(OutputFileName, ".WAV");	  // Add extension
      }

   // Ask signal level and MSX speed interface
	if (EncodeDecode == 1)
   {
		printf("\n  Specify MSX cassette interface speed: [1 - 1200bps, 2 - 2400bps] - ");
      ch = getch();
      if (ch == '2') MSXInterfaceSpeed = 2400;
      else if (ch == '1') MSXInterfaceSpeed = 1200;
      else ExitOnError(3);
      printf("%d", MSXInterfaceSpeed);
		printf("\n  Specify WAV signal level: [10 (lowest) - 99 (highest)] - ");
      ch = getch();
		SignalLevel = atoi(&ch);
      ch = getch();
      SignalLevel = SignalLevel * 10 + atoi(&ch);
      if (!SignalLevel || SignalLevel < 10) ExitOnError(3);
		printf("%d\n", SignalLevel);
      if (SignalLevel == 99) SignalLevel = 100;
   	}


   // Encode or decode file
   if (EncodeDecode == 1) EncodeFile();
   else if (EncodeDecode == 2) DecodeFile();

   // Cleanup
   CleanUp();

   // Final message
   printf("\n Operation completed successfully! Please send your comments to CasLink author:");
   printf("\n <podrezov@hotmail.com> or to ICQ UIN 1253313.\n\n");

	return 0;
	}


// Encode and write data byte to WAV file --------------------------------------

int WriteDataByte(char ch)
{
	int i, j;
   char BitMask = 1;

  	BitStream[0] = (char)STARTBIT;
  	BitStream[9] = (char)STOPBIT;	// Fill start and stopbits
  	BitStream[10] = (char)STOPBIT;
   for (i = 1; i < 9; i++)
   {
   	if (ch & BitMask) BitStream[i] = 1;	// Fill bit stream
      else BitStream[i] = 0;
      BitMask = BitMask << 1;
      }

   for (i = 0; i < sizeof(BitStream); i++)
   {
		if ((char)BitStream[i] == 0) for (j = 0; j < BitEncodingLength; j++) (unsigned char)OutputMemPointer[i * BitEncodingLength + j] = ZeroBit[SpeedScale][j];
      else for (j = 0; j < BitEncodingLength; j++) (unsigned char)OutputMemPointer[i * BitEncodingLength + j] = SetBit[SpeedScale][j];
      }

   // Save one encoded byte to file
	OutputFileLength = (size_t)(BitEncodingLength * sizeof(BitStream));
	BytesWritten = fwrite(OutputMemPointer, sizeof(char), OutputFileLength, OutputFileHandle);
	if (BytesWritten != OutputFileLength) return 1;

	return 0;
	}


// Abnormal termination routine ------------------------------------------------

void ExitOnError(int ErrorNum)
{
	// Clean up
   CleanUp();

   // Print error message and exit
   if (!ErrorNum) exit(255);
   printf("\n\n");
   printf(ErrorList[ErrorNum-1]);
   exit(ErrorNum);
   }


// Clean up routine ------------------------------------------------------------

void CleanUp(void)
{
	if (InputMemPointer) free(InputMemPointer);
	if (OutputMemPointer) free(OutputMemPointer);
   if (InputFileHandle) fclose(InputFileHandle);
   if (OutputFileHandle) fclose(OutputFileHandle);
   }



/*

 MSX Tape I/O
-------------

BIT AND BYTE ENCODING

Information on tape is stored as square waves of 1200 or 2400 baud. The
encoding used is FSK (frequency shifting key). The next table shows the
relation between the selected recording speed, the bits and the pulses
put on tape:

bps   bit   encoding          freq (Hz)
                    _______
1200   0    _______|       |  1200
                ___     ___
1200   1    ___|   |___|   |  2400

                ___
2400   0    ___|   |          2400
              _   _
2400   1    _| |_| |          4800

Each byte is put on tape as a sequence of 11 bits:

 0 b0 b1 b2 b3 b4 b5 b6 b7 1 1  (one startbit, 2 stopbits, LSb first)

Futhermore 2 kinds of headers are used: short and long, a long header being
4 times as long as a short one. They consist of a series of square wave pulses
(bit value 1):

bps   header   freq (Hz)   nr of pulses
1200  short    2400        4000             (5/3  sec)
1200  long     2400        16000            (20/3 sec)
2400  short    4800        8000
2400  long     4800        32000


FILE FORMATS

Files on tape are stored in two blocks. The first block is the filedescriptor,
the second contains the filedata.  Three types of files exist:
BASIC file, memory dump and data file.

BASIC File (CSAVE/CLOAD)
  filedescriptor: long header
                  10x #D3       (filetype)
                  6 bytes filename

  file data:      short header
                  tokenized BASIC
                  7x #00

memory dump (BSAVE/BLOAD)
  filedescriptor: long header
                  10x #D0
                  6 bytes filename

  file data:      short header
                  2 bytes start address
                  2 bytes end address
                  2 bytes exec address
                  memory dump

data file (SAVE/LOAD/OPEN)
  filedescriptor: long header
                  10x #EA
                  6 bytes filename

  file data:      short header
                  block of 256 bytes

                  short header
                  block of 256 bytes

                  ....
                  last block contains 256x #1A (^Z/EoF)

*/

// Decode file from WAV sound file routine -------------------------------------

void DecodeFile(void)
{

	}


// Encode file to WAV sound file routine ---------------------------------------

void EncodeFile(void)
{

   // Initialize variables
  	WavSampleRate = 11025;			// 11025, 22050, 44100 Hz (8-bit, mono)
   LongHeaderLength = 20/3;
   ShortHeaderLength = 5/3;
   FirstPauseLength = 2;			// Seconds
   MiddlePauseLength = 1;				// Seconds
   FinalPauseLength = 1;			// Seconds
	SilenceByte = SILENCE;


   // Prepare scales and bit patterns according to volume
   SampleScale = WavSampleRate / 11025;   // sample scale for different sampling rates
   SpeedScale = MSXInterfaceSpeed / 1200 - 1;	// speed scale for MSX interface
	if (!SpeedScale) BitEncodingLength = 10;
   else BitEncodingLength = 5;
   for (i = 0; i < BitEncodingLength; i++)
   {
		if (ZeroBit[SpeedScale][i] == (unsigned char)LOWAMPLITUDE) ZeroBit[SpeedScale][i] = (unsigned char)(LOWAMPLITUDE + (MAXSIGNALLEVEL - SignalLevel));
      else if (ZeroBit[SpeedScale][i] == (unsigned char)HIGHAMPLITUDE) ZeroBit[SpeedScale][i] = (unsigned char)(HIGHAMPLITUDE - (MAXSIGNALLEVEL - SignalLevel));
		if (SetBit[SpeedScale][i] == (unsigned char)LOWAMPLITUDE) SetBit[SpeedScale][i] = (unsigned char)(LOWAMPLITUDE + (MAXSIGNALLEVEL - SignalLevel));
      else if (SetBit[SpeedScale][i] == (unsigned char)HIGHAMPLITUDE) SetBit[SpeedScale][i] = (unsigned char)(HIGHAMPLITUDE - (MAXSIGNALLEVEL - SignalLevel));
    	}


   // Allocate memory for output buffer
   OutputMemPointer = (char*) malloc((size_t)(BitEncodingLength * sizeof(BitStream)));
   if (!OutputMemPointer) ExitOnError(4);


   // Open, check size, allocate memory and read input file
   InputFileHandle = fopen(InputFileName, "rb");
   if (!InputFileHandle) ExitOnError(6);
	if (fseek(InputFileHandle, 0, SEEK_END)) ExitOnError(7);
   InputFileLength = (size_t)ftell(InputFileHandle);
   if (fseek(InputFileHandle, 0, SEEK_SET)) ExitOnError(7);
	if (InputFileLength < 2L || InputFileLength > (long)MAXENCINPUTFILELENGTH) ExitOnError(8);
   InputMemPointer = (char*) malloc(InputFileLength);
   if (!InputMemPointer) ExitOnError(5);
   BytesRead = fread(InputMemPointer, sizeof(char), InputFileLength, InputFileHandle);
   if (BytesRead != InputFileLength) ExitOnError(9);


   // Check program type and make necessary adjustments
   if ((unsigned char)InputMemPointer[0] == 0xff)
   {
   	FileType = 1;			// tokenized basic program
      ExtraBytes = 7;
      FileOffset = 1;
      }
   else if ((unsigned char)InputMemPointer[0] == 0xfe)
   {
   	FileType = 2;			// BIN program
      ExtraBytes = 0;
      FileOffset = 1;
      }
   else if ((unsigned char)InputMemPointer[0] == 'A' && (unsigned char)InputMemPointer[1] == 'B')
   {
   	FileType = 3;			// ROM image
      ExtraBytes = 0;
      FileOffset = 0;
		}
   else
   {
		for (i = 0; i < InputFileLength; i++) if ((unsigned char)InputMemPointer[i] < ' ') ExitOnError(10);
   	FileType = 4;			// ASCII basic file
      ExtraBytes = 256;
      FileOffset = 0;
   	}


   // Fill WAV header
	strcpy(WavHeader.WavHeaderID, "RIFF");
	strcpy(WavHeader.MMFileType, "WAVE");
	strcpy(WavHeader.FormatID, "fmt ");
  	strcpy(WavHeader.DataID, "data");
   WavHeader.Dummy = (WORD) 1;
   WavHeader.FormatTag = (WORD) 1;	// PCM
   WavHeader.NumChannels = (WORD) 1; // mono always
   WavHeader.BitsPerSample = (WORD) 8;	// 8 bit always
   WavHeader.SamplesPerSec = (DWORD)WavSampleRate;
   WavHeader.BytesPerSec = (DWORD)((DWORD)WavSampleRate * (DWORD)(WavHeader.BitsPerSample / 8) * (DWORD)WavHeader.NumChannels);
	WavHeader.FormatSize = (DWORD)(sizeof(WavHeader.FormatTag) + sizeof(WavHeader.NumChannels) + sizeof(WavHeader.BitsPerSample) +
   	sizeof(WavHeader.SamplesPerSec) + sizeof(WavHeader.BytesPerSec) + sizeof(WavHeader.Dummy));
	if (FileType == 1) FileHeaderIDLength = sizeof(BasicFileHeaderID);
   else if (FileType == 2 || 	FileType == 3) FileHeaderIDLength = sizeof(BinFileHeaderID);
	if (FileType == 4) FileHeaderIDLength = sizeof(DataFileHeaderID);
   WavHeader.PureSampleLength = (DWORD)((DWORD)WavSampleRate * (DWORD)FirstPauseLength + (DWORD)WavSampleRate * (DWORD)MiddlePauseLength + (DWORD)WavSampleRate * (DWORD)FinalPauseLength) +	// Length of pauses
   	(DWORD)((DWORD)WavSampleRate * (DWORD)LongHeaderLength + (DWORD)WavSampleRate * (DWORD)ShortHeaderLength) +	// Length of long and short headers
   		(DWORD)(((DWORD)FileHeaderIDLength + (DWORD)CASFILENAMELENGTH + (DWORD)InputFileLength + (DWORD)ExtraBytes - (DWORD)FileOffset) * (DWORD)SampleScale * (DWORD)sizeof(BitStream) * (DWORD)BitEncodingLength);	// Length of data areas

   WavHeader.PureSampleLength = WavHeader.PureSampleLength - 5;
   WavHeader.SampleLength = (DWORD)((DWORD)WavHeader.PureSampleLength + (DWORD)WAVHEADEROFFSETVALUE);


   // Check if file already exitis
   OutputFileHandle = fopen(OutputFileName, "rb");
	if (OutputFileHandle)
   {
	   printf("\n WARNING: File with this name already exists. Overwrite ? [Y/N] - ");
      ch=getch();
      if (toupper(ch) == 'N') ExitOnError(11);
      printf("%c\n", toupper(ch));
	   fclose(OutputFileHandle);
      }

   // Write output file header (WAV)
   OutputFileHandle = fopen(OutputFileName, "wb");
   if (!OutputFileHandle) ExitOnError(12);
   BytesWritten = fwrite(&WavHeader, sizeof(char), sizeof(WavHeader), OutputFileHandle);
   if (BytesWritten != sizeof(WavHeader)) ExitOnError(13);


   // Write first pause
	for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)FirstPauseLength); j++)
   {
	   BytesWritten = fwrite(&SilenceByte, sizeof(char), sizeof(SilenceByte), OutputFileHandle);
	   if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
      }


   // Create and write long header
	for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)LongHeaderLength / (DWORD)BitEncodingLength); j++)
   {
	   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(char), BitEncodingLength, OutputFileHandle);
   	if (BytesWritten != BitEncodingLength) ExitOnError(13);
      }


	// Prepare file name for tape header
   for (i = 0; i < CASFILENAMELENGTH; i++)
   {
   	if ((char)InputFileName[i] != '.') (char)NameBuffer[i] = (char)InputFileName[i];
      else break;
      }
   for ( ; i < CASFILENAMELENGTH; i++) (char)NameBuffer[i] = ' ';


   // Write file header data
	if (FileType == 1) for (i = 0; i < sizeof(BasicFileHeaderID); i++) if (WriteDataByte(BasicFileHeaderID[i])) ExitOnError(13);	// Basic ID
   else if (FileType == 2 || FileType == 3) for (i = 0; i < sizeof(BinFileHeaderID); i++) if (WriteDataByte(BinFileHeaderID[i])) ExitOnError(13);	// BIN/ROM ID
   else if (FileType == 4) for (i = 0; i < sizeof(DataFileHeaderID); i++) if (WriteDataByte(DataFileHeaderID[i])) ExitOnError(13);	// Data ID
	for (i = 0; i < CASFILENAMELENGTH; i++) if (WriteDataByte(NameBuffer[i])) ExitOnError(13);	// File name


   // Write second pause
	for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)MiddlePauseLength); j++)
   {
	   BytesWritten = fwrite(&SilenceByte, sizeof(char), sizeof(SilenceByte), OutputFileHandle);
	   if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
      }


   // Create and write short header
	for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)ShortHeaderLength / (DWORD)BitEncodingLength); j++)
   {
	   BytesWritten = fwrite(SetBit[SpeedScale], sizeof(char), BitEncodingLength, OutputFileHandle);
   	if (BytesWritten != BitEncodingLength) ExitOnError(13);
      }


   // Write program data
	for (i = FileOffset; i < InputFileLength; i++)
   {
   	ch = (char)InputMemPointer[i];	// Retrieve data byte
      if (WriteDataByte(ch)) ExitOnError(13);	// Write data byte
      }

	if (FileType == 1) for (i = 0; i < 7; i++) if (WriteDataByte(0)) ExitOnError(13);	// Write 7 zeros at the end of basic program
	else if (FileType == 4) for (i = 0; i < 256; i++) if (WriteDataByte(0x1a)) ExitOnError(13);	// Write 256 bytes at the end of ascii program


   // Write final pause
	for (j = 0; j < (DWORD)((DWORD)WavSampleRate * (DWORD)FinalPauseLength); j++)
   {
	   BytesWritten = fwrite(&SilenceByte, sizeof(char), sizeof(SilenceByte), OutputFileHandle);
	   if (BytesWritten != sizeof(SilenceByte)) ExitOnError(13);
      }

   return;
	}